"""Functions for focus stacking of images."""

import numpy as np
from scipy import ndimage
import pywt
import dtcwt


def merge_dtcwt(first, second, levels=None, sigma=None):
    """Focus-stack two images using the dual-tree complex wavelet
       transform (DTCWT).

    Parameters
    ----------
    first : array_like
        First input image.
    second : array_like
        Second input image.
    levels : int
        Number of levels to use in the DTCWT.
    sigma : float
        Standard deviation of lowpass filter used to smooth the mixing weights.

    Returns
    -------
    stacked : numpy.ndarray
        Focus-stacked result.
    """
    if levels is None:
        levels = 3
    if sigma is None:
        sigma = 1.5

    transform = dtcwt.Transform2d()

    def merge_channel(fc, sc):
        fc = transform.forward(fc, nlevels=levels)
        sc = transform.forward(sc, nlevels=levels)
        fc.lowpass[:] = 0.5 * (fc.lowpass + sc.lowpass)
        for f, s in zip(fc.highpasses, sc.highpasses):
            mix = 1.0 * (np.abs(f) > np.abs(s))
            mix = np.median(mix, axis=2)
            mix = ndimage.gaussian_filter(mix, sigma=sigma)
            mix = mix[:, :, np.newaxis]
            f[:] = mix * f + (1 - mix) * s
        return transform.inverse(fc)

    if first.ndim == 2:
        return merge_channel(first, second)

    return np.dstack([
        merge_channel(first[:, :, c], second[:, :, c])
        for c in range(first.ndim)
    ])


def merge_highpass(first, second, sigma=None):
    """Focus-stack two images using a Gaussian highpass filter.

    Parameters
    ----------
    first : array_like
        First input image.
    second : array_like
        Second input image.
    sigma : float
        Standard deviation of the highpass filter.

    Returns
    -------
    stacked : numpy.ndarray
        Focus-stacked result.
    """
    if sigma is None:
        sigma = 1.5

    def sharpness(image):
        bw = image
        if image.ndim == 3:
            bw = np.mean(bw, axis=2)
        lowpass = ndimage.gaussian_filter(bw, sigma=sigma)
        highpass = bw - lowpass
        return ndimage.gaussian_filter(np.abs(highpass), sigma=sigma)

    first_sharpness = sharpness(first)
    second_sharpness = sharpness(second)
    diff = first_sharpness - second_sharpness
    mix = 1 / (1 + np.exp(-100 * diff))
    if first.ndim == 3:
        mix = mix[:, :, np.newaxis]
    return mix * first + (1 - mix) * second


def merge_waveletes(first, second, levels=None, sigma=None):
    """Focus-stack two images using a wavelet transform.

    Parameters
    ----------
    first : array_like
        First input image.
    second : array_like
        Second input image.
    sigma : float
        Standard deviation of lowpass filter used to smooth the mixing weights.

    Returns
    -------
    stacked : numpy.ndarray
        Focus-stacked result.
    """
    if levels is None:
        levels = 4
    if sigma is None:
        sigma = 1.5

    rgb = first.ndim == 3
    wavelet = 'db6'
    first = pywt.wavedec2(first, wavelet, axes=(0, 1))
    second = pywt.wavedec2(second, wavelet, axes=(0, 1))

    coeffs = [0.5 * (first[0] + second[0])]
    for first_cs, second_cs in zip(first[1:], second[1:]):
        merged_cs = []
        for first_c, second_c in zip(first_cs, second_cs):
            mix = 1.0 * (np.abs(first_c) > np.abs(second_c))
            if rgb:
                mix = np.mean(mix, axis=2)
            mix = ndimage.gaussian_filter(mix, sigma=sigma)
            if rgb:
                mix = mix[:, :, np.newaxis]
            merged_cs.append(mix * first_c + (1 - mix) * second_c)
        coeffs.append(merged_cs)

    return np.maximum(pywt.waverec2(coeffs, wavelet, axes=(0, 1)), 0)


def merge_waveletes2(first, second, levels=None):
    """Focus-stack two images using a wavelet transform.

    Parameters
    ----------
    first : array_like
        First input image.
    second : array_like
        Second input image.
    sigma : float
        Standard deviation of lowpass filter used to smooth the mixing weight.

    Returns
    -------
    stacked : numpy.ndarray
        Focus-stacked result.
    """
    if levels is None:
        levels = 4

    wavelet = 'db6'
    rgb = first.ndim == 3

    first_coeffs = pywt.wavedec2(np.mean(first, axis=2) if rgb else first,
                                 wavelet,
                                 level=levels,
                                 axes=(0, 1))[1:]
    second_coeffs = pywt.wavedec2(np.mean(second, axis=2) if rgb else second,
                                  wavelet,
                                  level=levels,
                                  axes=(0, 1))[1:]

    mix = np.full(first.shape[:2], 0.5, dtype='float32')

    for f, s in zip(reversed(first_coeffs), reversed(second_coeffs)):
        f = sum(np.abs(fi) for fi in f)
        s = sum(np.abs(si) for si in s)
        f = ndimage.zoom(
            f, (first.shape[0] / f.shape[0], first.shape[1] / f.shape[1]))
        s = ndimage.zoom(
            s, (first.shape[0] / s.shape[0], first.shape[1] / s.shape[1]))
        undefined = mix == 0.5

        fbig = np.bitwise_and(undefined, f > 2 * s)
        sbig = np.bitwise_and(undefined, s > 2 * f)
        mix[fbig] = 1
        mix[sbig] = 0

    if rgb:
        mix = mix[:, :, np.newaxis]
    return mix * first + (1.0 - mix) * second



"""Command line tool of rawhdr."""

import os

import click
import imageio
import rawpy

import rawhdr


def print_version(ctx, _, value):
    """Print version information."""
    if not value or ctx.resilient_parsing:
        return
    click.echo('rawhdrmerge version ' + rawhdr.__version__ + '\n'
               'Copyright (C) 2019-2020 Felix Thaler')
    ctx.exit()


def load_image(path):
    """Load a raw image file.

    Parameters
    ----------
    path : str
        Path to the image file.

    Returns
    -------
    image : ndarray
        Loaded image data in linear color space.
    """
    with rawpy.imread(path) as raw:
        rgb = raw.postprocess(gamma=(1, 1),
                              no_auto_bright=True,
                              use_camera_wb=True,
                              output_bps=16)
    return rgb.astype('float32') / 2**16


@click.group()
@click.option('--version',
              '-v',
              is_flag=True,
              callback=print_version,
              expose_value=False,
              is_eager=True)
def main():
    pass


@main.command()
@click.argument('images', nargs=-1, type=click.Path(exists=True))
@click.option('--output',
              '-o',
              type=click.Path(),
              help='File name of the output HDR image.')
@click.option('--save-memory',
              '-s',
              is_flag=True,
              help='Use a less memory-intense algorithm that could lead '
              'to slight quality reduction in some cases.')
@click.option('--mask-width',
              type=float,
              help='Mask width option used for exposure scaling.')
@click.option('--blend-width',
              type=float,
              help='Smoothness option used for blending.')
@click.option('--blend-cap', type=float, help='Cap option used for blending.')
@click.option('--target-gamma',
              type=float,
              help='Gamma correction used in internal computations.')
def hdr_merge(images, output, save_memory, mask_width, blend_width, blend_cap,
              target_gamma):
    """Command-line utility for merging RAW images into a single HDR image.

    All input images must be RAW images. The exposure of the first image is
    taken as reference for the brightness of the resulting HDR image.
    """
    from rawhdr import merge

    if not images:
        return

    if output is None:
        output = os.path.splitext(images[0])[0] + '-hdr.exr'

    if save_memory:
        # Load one image after another to save memory
        merged, other_images = images[0], images[1:]
        merged = load_image(merged)
        for image in other_images:
            image = load_image(image)
            merged = merge.merge_exposures([merged, image],
                                           mask_width=mask_width,
                                           blend_width=blend_width,
                                           blend_cap=blend_cap,
                                           target_gamma=target_gamma,
                                           weight_first=False)
            del image
    else:
        # Load all images at ones and perform merging
        images = [load_image(image) for image in images]
        merged = merge.merge_exposures(images,
                                       mask_width=mask_width,
                                       blend_width=blend_width,
                                       blend_cap=blend_cap,
                                       target_gamma=target_gamma)

    imageio.imsave(output, merged.astype('float32'))


@main.command()
@click.argument('images', nargs=-1, type=click.Path(exists=True))
@click.option('--output',
              '-o',
              type=click.Path(),
              help='File name of the stacked output image.')
@click.option('--sigma',
              type=float,
              help='Standard deviation of filter used in merger '
              '(availabilty and exact meaning depends on --merger option).')
@click.option('--levels',
              type=int,
              help='Number of levels used in wavelet transforms '
              '(availabilty and exact meaning depends on --merger option).')
@click.option('--merger',
              type=click.Choice(['highpass', 'wavelet', 'dtcwt', 'wavelet2']),
              help='Kind of merger to use.',
              default='highpass')
def focus_stack(images, output, sigma, levels, merger):
    """Command-line utility for focus stacking of RAW images.

    All input images must be RAW images.
    """
    from rawhdr import focusstack

    if not images:
        return

    if output is None:
        output = os.path.splitext(images[0])[0] + '-stacked.exr'

    def merge(first, second):
        if merger == 'dtcwt':
            return focusstack.merge_dtcwt(first,
                                          second,
                                          levels=levels,
                                          sigma=sigma)
        if merger == 'highpass':
            if levels is not None:
                raise ValueError(
                    'Option --levels is not supported for this merger.')
            return focusstack.merge_highpass(first, second, sigma=sigma)
        elif merger == 'wavelet':
            return focusstack.merge_waveletes(first,
                                              second,
                                              levels=levels,
                                              sigma=sigma)
        elif merger == 'wavelet2':
            if sigma is not None:
                raise ValueError(
                    'Option --sigma is not supported for this merger.')
            return focusstack.merge_waveletes2(first, second, levels=levels)
        else:
            raise ValueError()

    images = list(images)
    stacked = load_image(images.pop())
    while images:
        image = load_image(images.pop())
        stacked = merge(stacked, image)

    imageio.imsave(output, stacked.astype('float32'))


if __name__ == '__main__':
    main()



"""Setup file for rawhdr."""

import os
import re
import setuptools


def read_file(*path):
    """Read file content."""
    package_path = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(package_path, *path), 'r') as text_file:
        return text_file.read()


def long_description():
    """Read long description from README.md."""
    return read_file('README.md')


def version():
    """Parse version info."""
    initfile_content = read_file('rawhdr', '__init__.py')
    match = re.search(r"__version__ = [']([^']*)[']", initfile_content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError('Unable to find version string')


setuptools.setup(name='rawhdr',
                 version=version(),
                 author='Felix Thaler',
                 author_email='felix.thaler@nummi.ch',
                 description='A simple HDR image merger',
                 long_description=long_description(),
                 long_description_content_type='text/markdown',
                 url='https://github.com/fthaler/rawhdr',
                 packages=setuptools.find_packages(),
                 entry_points={'console_scripts': ['rawhdr=rawhdr.cli:main']},
                 install_requires=[
                     'click', 'dtcwt', 'imageio', 'numpy', 'PyWavelets',
                     'rawpy', 'scipy'
                 ],
                 classifiers=[
                     'Programming Language :: Python',
                     'License :: OSI Approved :: '
                     'GNU General Public License v2 or later (GPLv2+)',
                     'Operating System :: OS Independent',
                     'Topic :: Multimedia :: Graphics :: Graphics Conversion'
                 ])



"""Unit tests of rawhdr.focusstack."""

import unittest

import numpy as np
from scipy import ndimage

from rawhdr import focusstack


def _generate_images():
    n = 200
    sharp = np.random.RandomState(42).uniform(size=(n, 300, 3))
    image0 = np.copy(sharp)
    image1 = np.copy(sharp)
    image0[n // 2:, :, :] = 0.5
    image1[:n // 2, :, :] = 0.5
    return sharp, image0, image1


_MERGERS = [
    focusstack.merge_dtcwt, focusstack.merge_highpass,
    focusstack.merge_waveletes, focusstack.merge_waveletes2
]


def _sharpness(image):
    bw = image
    if image.ndim == 3:
        bw = np.mean(bw, axis=2)
    lowpass = ndimage.gaussian_filter(bw, sigma=1.5)
    highpass = bw - lowpass
    return ndimage.gaussian_filter(np.abs(highpass), sigma=1.5)


class TestStackImages(unittest.TestCase):
    def mergers(self):
        for merge in _MERGERS:
            with self.subTest(merge=merge.__name__):
                yield merge

    """Tests for focusstack.stack_images()."""

    def test_equal(self):
        """Test stacking of images with equal sharpness."""
        for merge in self.mergers():
            sharp, _, _ = _generate_images()
            stacked = merge(sharp, sharp)
            self.assertTrue(np.allclose(stacked, sharp))

    def test_improved_sharpness(self):
        """Test sharpness improvement of stacking."""
        _, image0, image1 = _generate_images()
        for merge in self.mergers():
            stacked = merge(image0, image1)

            self.assertGreater(np.mean(_sharpness(stacked)),
                               np.mean(_sharpness(image0)))
            self.assertGreater(np.mean(_sharpness(stacked)),
                               np.mean(_sharpness(image1)))

    def test_close_to_sharp(self):
        """Check that stacked image sharpness is close to perfectly sharp."""
        for merge in self.mergers():
            sharp, image0, image1 = _generate_images()
            stacked = merge(image0, image1)
            self.assertGreater(
                np.mean(_sharpness(stacked)) / np.mean(_sharpness(sharp)), 0.9)

    def test_bw_input(self):
        for merge in self.mergers():
            rng = np.random.RandomState(42)
            first = rng.uniform(size=(200, 300))
            second = rng.uniform(size=(200, 300))
            stacked = merge(first, second)
            self.assertEqual(stacked.shape, (200, 300))
